"""Django app providing ``HashableFileFieldsBaseModel``, a base django model to automatically compute
file-fields hashes."""

from os import path

import pkg_resources
from setuptools.config import read_configuration


def _extract_version(package_name):
    """Extract the installed version of the given package name.

    Parameters
    ----------
    package_name : str
        The package for which we want the version

    Returns
    -------
    str
        The installed version as a string

    """
    try:
        # if package is installed
        version = pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        # if not installed, so we must be in source, with ``setup.cfg`` available
        _conf = read_configuration(path.join(path.dirname(__file__), "..", "setup.cfg"))
        version = _conf["metadata"]["version"]

    return version


__version__ = _extract_version("django_hashable_file_fields")
VERSION = tuple(int(part) for part in __version__.split(".") if str(part).isnumeric())



"""Simple hashers and utils for the ``django-hashable-file-fields`` app."""

import hashlib


def md5_hash(file_field, instance):  # pylint: disable=unused-argument
    """An example hasher that uses ``hashlib.md5`` to compute the hash of a file field.

    Parameters
    ----------
    file_field : django.db.models.fields.files.FileField
        The file field (or img...) for which we want to compute the hash.
    instance : HashableFileFieldsBaseModel
        The instance holding this field.

    Returns
    -------
    str
        The computed hash for the given file field.


    Example
    -------

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     img = models.ImageField(...)
    ...     hash = models.CharField(...)
    ...
    ...     hashable_fields = [
    ...         ('img', 'hash', md5_hash)
    ...     ]

    """

    md5 = hashlib.md5()
    for chunk in file_field.chunks():
        md5.update(chunk)
    return md5.hexdigest()


def with_default(default, hash_function=None, exceptions=None):
    """Decorator to a hasher that will return a default value when an exception occurs.

    Parameters
    ----------
    default : ?
        The default value to return as a hash value when an exception is raised during the call
        to ``hash_function``.

    hash_function : function, optional
        The function that takes a file field, its parent model instance, and return a computed
        hash of this field.
        If not defined, the default function defined on the model at call time will be used.

    exceptions : iterable, optional
        A list (or other iterable) of exceptions classes for which to return the value defined in
        ``default`` when an exception is raised during the call to the ``hash_function``.
        Other exceptions will be raised.
        By default all exceptions based on``Exception`` are handled.

    Returns
    -------
        Either the value returned by calling ``hash_function``, or the default value if a valid
        exception is raised.

    Example
    -------

    Using the default hash function:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     img = models.ImageField(...)
    ...     hash = models.CharField(...)
    ...
    ...     hashable_fields = [
    ...         ('img', 'hash', with_default(settings.EMPTY_FILE_HASH))
    ...     ]

    Using a specific hash function:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     img = models.ImageField(...)
    ...     hash = models.CharField(...)
    ...
    ...     hashable_fields = [
    ...         ('img', 'hash', with_default(settings.EMPTY_FILE_HASH, md5_hash))
    ...     ]

    """

    if exceptions is not None:
        exceptions = tuple(exceptions)
    else:
        exceptions = (Exception, )

    def func(file_field, instance):
        """Calls the hash function and returns the default value in case of exception."""
        try:
            final_hash_function = hash_function

            # Use the default hash function if none was passed to ``with_default``
            if not final_hash_function:
                # If we get it on the instance using ``instance.default_hash_function``, we have a
                # bound method. But we really want the original function not taking ``self`` as
                # first argument so we get it on the class.
                final_hash_function = instance.__class__.default_hash_function

            return final_hash_function(file_field, instance)

        except exceptions:  # pylint: disable=catching-non-exception
            return default

    return func



"""Provides ``Hashable``, a base django model for the ``django-hashable-file-fields`` app.

The simplest usage is:

>>> class MyModel(HashableFileFieldsBaseModel):
...     img = models.ImageField(...)
...     hash = models.CharField(...)
...
...     hashable_fields = [
...         ('img', 'hash')
...     ]

It will save the md5 hash value of the ``img`` field in the ``hash`` field on each save, raising
on exception.

"""

from django.db import models

from .hashers import md5_hash


class HashableFileFieldsBaseModel(models.Model):
    """A base model to easily compute and store hashed values of file fields.

    Attributes
    ----------
    hashable_fields : list
        The list of all fields to hash. This may be the only thing to configure in models that
        inherit from ``HashableFileFieldsBaseModel``.
        Each entry is a tuple composed of two or three elements:
        1. The name of the file field to hash
        2. The name of the file that will contain the hash
        3. The function to use to compute the hash. If not set, the function defined in the
           ``default_hash_function`` attribute will be used (default to ``md5_hash`` from
            ``hashable.hashers``.

    default_hash_function : str
        The function used to compute the hash of a field if the third element of a tuple defined in
        ``hashable_fields`` is not set.
        This function takes a file field and its parent model instance, and return a computed
        hash of this field.

    compute_hash_method_pattern : str
        Instead of using an external hash function (or use ``default_hash_function``, it's
        possible to use a method of the model to compute the hash for a field. This function
        must follow the pattern defined in ``compute_hash_method_pattern``, which is a formatable
        string where ``%(field_name)s`` and ``%(hash_field_name)s`` will be replaced before
        calling it.
        The method must accept the name of the field to hash, and optionally an external hash
        function and must return the hashed value to save.
        The default pattern is ``'compute_hash_%(field_name)s'`` so if the file to hash is named
        ``img``, the method must be named ``compute_hash_img``.

    default_compute_hash_method : function
        The method that will be used if no methd exists following the pattern defined in
        ``compute_hash_method_pattern``.

    Example
    -------
    The minimal configuration to use this ``HashableFileFieldsBaseModel`` model is to simply
    inherit from it and define the name of the field to hash and the field in which to save the
    hash, this way:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     img = models.ImageField(...)
    ...     hash = models.CharField(...)
    ...
    ...     hashable_fields = [
    ...         ('img', 'hash')
    ...     ]

    With this, all the default configuration will be used:
    - the ``compute_hash`` default method will be called
    - this method will call ``md5_hash`` function, raising if something bad happen in it.

    To save a default hash field in the hash function raise, you can define the
    ``default_hash_function`` this way:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     # ...
    ...     default_hash_function = with_default(settings.EMPTY_FILE_HASH, md5_hash))

    Or, for a specific field:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     # ...
    ...     hashable_fields = [
    ...         ('img', 'hash'),
    ...         ('img2', 'hash2', with_default(settings.EMPTY_FILE_HASH, md5_hash))
    ...         # ``with_default`` without hash function will use the default one
    ...         ('img3', 'hash3', with_default(settings.EMPTY_FILE_HASH))
    ...     ]

    OR it can be defined with a method:

    >>> class MyModel(HashableFileFieldsBaseModel):
    ...     # ...
    ...     hashable_fields = [
    ...         ('img', 'hash'),
    ...         ('img2', 'hash2')
    ...     ]
    ...
    ...     def compute_hash_img2(self, field_name, hash_method=None):
    ...         return do_some_computation_with_field_name_and_maybe_hash_method()

    """

    hashable_fields = []
    default_hash_function = md5_hash
    compute_hash_method_pattern = 'compute_hash_%(field_name)s'
    default_compute_hash_method = 'compute_hash'

    class Meta:
        abstract = True

    def compute_hash(self, field_name, hash_function=None):
        """Default method that will call a hash_function to compute the hash of a file field.

        Parameters
        ----------
        fied_name : str
            The name of the file field to be hashed

        hash_function : function, optional
            The function that takes a file field, its parent model instance, and return a computed
            hash of this field.
            If not set, the function defined in ``default_hash_function`` will be used.
            A "hash function" expect the file field of an instance (ie ``self.img`` if ``img`` is
            a ``ImageField``), and the instance itself.


        Returns
        -------
            The result of the hash function.

        """

        if hash_function is None:
            # If we get it on the instance using ``self.default_hash_function``, we have a bound
            # method. But we really want the original function not taking ``self`` as first
            # argument so we get it on the class.
            hash_function = self.__class__.default_hash_function

        if not hash_function:
            raise RuntimeError('Cannot call default compute hash method without '
                               '`hash_function`')

        return hash_function(getattr(self, field_name), self)

    def compute_hashable_fields(self):
        """Compute, and set, all the hash defined in the ``hashable_fields`` attribute.

        For each entry of ``hashable_fields`` it will check if a method matching
        ``compute_hash_method_pattern`` exists and if not, will use ``compute_hash``.
        Then this method will be called to get the hash, and this hash will be set in ``self``.
        The current instance will *not* be saved by this method.

        """
        for hash_infos in self.hashable_fields:

            field_name, hash_field_name = hash_infos[:2]

            # We may not have defined ``hash_function`` in the tuple
            compute_hash_function = hash_infos[2] if len(hash_infos) == 3 else None

            # Get the method to call based on the field name and hash field name.
            hash_method_name = self.compute_hash_method_pattern % {
                'field_name': field_name,
                'hash_field_name': hash_field_name,
            }

            # The default compute method will be used if this specific method doesn't exist.
            hash_method = getattr(self, hash_method_name,
                                  getattr(self, self.default_compute_hash_method))

            # Call the method, and set the value in the hash field
            setattr(self, hash_field_name, hash_method(field_name, compute_hash_function))

    def save(self, *args, **kwargs):
        """Compute the hash fields then call ``super`` to save the instance."""

        self.compute_hashable_fields()

        return super().save(*args, **kwargs)
